from .DefDict import DefDict
from .UnitType import UnitType
from .MapRule import MapRule
from .StdUnit import *
import SpeedTest


__author__ = 'UCLA LEMUR'
__credits__ = 'The Laboratory for Embedded Machines and Ubiquitous Robots'
__url__ = 'https://github.com/Suke0811/DefDict/'
__description__ = 'Defined Dictionary'
